# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Copyright (c) 2000-2003 LOGILAB S.A. (Paris, FRANCE).
 http://www.logilab.fr/ -- mailto:contact@logilab.fr

utilities methods and classes for checkers
"""

__revision__ = "$Id: __init__.py,v 1.6 2003/10/04 15:33:09 syt Exp $"

import tokenize
from os import listdir
from os.path import join, isdir, splitext

from logilab.common.astng import ASTNGHandler
from logilab.common.configuration import OptionsProviderMixIn


class CheckerHandler:
    """implements IChecker methods
    """
    
    def set_option(self, option, opt, value, parser, *args, **kwargs):
        """call back method called to set an option
        """

    def open(self):
        """called before visiting project (i.e set of modules)"""
        
    def close(self):
        """called after visiting project (i.e set of modules)"""

    def open_module(self):
        """called before visiting a module"""
        
    def close_module(self):
        """called after visiting a module"""
        
    def get_layouts(self):
        """return a list of layouts objects to control results display"""
        return []

    
class ASTNGCheckerHandler(CheckerHandler, ASTNGHandler):
    """implements IASTNGChecker methods
    """

    
class BaseChecker(OptionsProviderMixIn):
    """base class for checkers"""

    options = ()
    priority = -9
    may_be_disabled = 1
    name = None
    
    def __init__(self, linter=None):
        """checker instances should have the linter as argument

        linter is an object implementing ILinter
        """
        self.name = self.name.lower()
        if self.may_be_disabled:
            opt_name = 'enable-' + self.name
            self.options = (
                (opt_name,
                 {'type' : 'yn', 'default' : 1, 'metavar': '<y_or_n>',
                  'help' : "enable/disable this checker"})
                ,) + self.options            
        OptionsProviderMixIn.__init__(self)
        self.linter = linter
        self._layouts = []
        
    def add_message(self, type, msg, line=None, node=None):
        """add a message of a given type"""
        self.linter.add_message(type, msg, line, node)

    def get_layouts(self):
        """return previously built layouts"""
        return self._layouts

    def is_enabled(self):
        """return true if the checker is enabled"""
        opt = 'enable_' + self.name
        return getattr(self.config, opt, 1)

    def enable(self, enable):
        """enable / disable this checker if true / false is given

        it false values has no effect if the checker can't be disabled
        """
        if self.may_be_disabled:
            setattr(self.config, 'enable_' + self.name, enable)
        
class BaseRawChecker(BaseChecker):
    """base class for raw checkers"""
    
    def process_module(self, name, stream):
        """process the module with the given name
        
        the module's content is accessible via the stream object
        
        stream must implements the readline method
        """
        try:
            self.process_tokens(tokenize.generate_tokens(stream.readline))
        except tokenize.TokenError, msg:
            self.add_message('fatal','Token Error in %s (%s)' % (file, msg))
    
    def process_tokens(self, tokens):
        """should be overiden by syb classes"""
        raise NotImplementedError()

PYTHON_EXTENSIONS = ('.py', '.pyc', '.pyo', '.so')

def initialize(linter):
    """initialize linter with checkers in this package """
    package_load(linter, __path__[0])

def package_load(linter, directory):
    """load all module and package in the given directory, looking for a
    'register' function in each one, used to register pylint checkers
    """
    g = globals()
    imported = {}
    for file in listdir(directory):
        basename, extension = splitext(file)
        if not imported.has_key(basename) and ((
            extension in PYTHON_EXTENSIONS and basename != '__init__') or (
            not extension and not basename == 'CVS' and
            isdir(join(directory, basename)))):
            try:
                ns = __import__(basename, g, g, None)
                if hasattr(ns, 'register'):
                    ns.register(linter)
                    imported[basename] = 1
            except ImportError:
                import sys
                print >> sys.stderr, "Problem importing module: %s" % file

__all__ = ('CheckerHandler', 'ASTNGCheckerHandler', 'BaseChecker',
           'initialize', 'package_load')
