# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Copyright (c) 2000-2003 LOGILAB S.A. (Paris, FRANCE).
 http://www.logilab.fr/ -- mailto:contact@logilab.fr

 basic checker for Python code
"""

__revision__ = "$Id: base.py,v 1.17 2003/10/09 14:50:52 alf Exp $"

from logilab.common.astng import ImportStmt, GlobalStmt

from logilab.pylint.interfaces import IASTNGChecker, IStatable
from logilab.pylint.checkers import BaseChecker, ASTNGCheckerHandler
from logilab.pylint.reporters import diff_string
from logilab.pylint.layouts import SectionLayout, TableLayout


import re

# regex for class/function/variable/constant nane
CLASS_NAME_RGX = re.compile('[A-Z][a-zA-Z1-9]+')
FUNC_NAME_RGX = re.compile('[a-z1-9_]+')
VAR_NAME_RGX = MOD_NAME_RGX = re.compile('[a-z1-9_]+')
CST_NAME_RGX = re.compile('[A-Z1-9_]+')

# do not require a doc string on system methods
NO_REQUIRED_DOC_RGX = re.compile('__.*__')

del re

def in_nested_list(nested_list, object):
    """return true if the object is an element of <nested_list> or of a nested
    list
    """
    for elmt in nested_list:
        if type(object) is type([]):
            if in_nested_list(nested_list, object):
                return 1
        if elmt == object:
            return 1
    return 0
    
class BasicChecker(BaseChecker, ASTNGCheckerHandler):
    """checks for :                                                            
    * doc strings                                                              
    * modules / classes / functions / methods / arguments / variables name     
    * number of arguments, local variables, branchs, returns and statements in
functions, methods                                                       
    * required module attributes                                             
    * dangerous default values as arguments                                    
    * redefinition of function / method / class                                
    * uses of the global statement                                             
    """
    
    __implements__ = (IASTNGChecker, IStatable)

    name = 'basic'
    priority = -1
    options = (('max-args',
                {'default' : 5, 'type' : 'int', 'metavar' : '<int>',
                 'help': 'Maximum number of arguments for function / method'}
                ),
               ('max-locals',
                {'default' : 15, 'type' : 'int', 'metavar' : '<int>',
                 'help': 'Maximum number of locals for function / method body'}
                ),
               ('max-returns',
                {'default' : 12, 'type' : 'int', 'metavar' : '<int>',
                 'help': 'Maximum number of return / yield for function / '
                         'method body'}
                ),
               ('max-branchs',
                {'default' : 12, 'type' : 'int', 'metavar' : '<int>',
                 'help': 'Maximum number of branch for function / method body'}
                ),
               ('max-statements',
                {'default' : 50, 'type' : 'int', 'metavar' : '<int>',
                 'help': 'Maximum number of statements in function / method '
                         'body'}
                ),
               ('required-attributes',
                {'default' : ('__revision__',), 'type' : 'csv',
                 'metavar' : '<attributes>',
                 'help' : 'Required attributes for module, separated by a '
                          'comma'}
                ),
               ('global-warning',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help' : 'Warning on global statements'}
                ),
               ('unreachable',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help' : 'Check for unreachable code'}
                ),
               ('redefinition',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help' : 'Check for function / method / class redefinition'}
                ),
               ('no-docstring-rgx',
                {'default' : NO_REQUIRED_DOC_RGX,
                 'type' : 'regexp', 'metavar' : '<regexp>',
                 'help' : 'Regular expression which should only match '
                          'functions which do not require a docstring'}
                ),
               ('min-name-length',
                {'default' : 3, 'type' : 'int', 'metavar' : '<int>',
                 'help': 'Minimal length for module / class / function / '
                         'method / argument / variable names'}
                ),
               ('module-rgx',
                {'default' : MOD_NAME_RGX,
                 'type' :'regexp', 'metavar' : '<regexp>',
                 'help' : 'Regular expression which should only match correct '
                          'module names'}
                ),
               ('class-rgx',
                {'default' : CLASS_NAME_RGX,
                 'type' :'regexp', 'metavar' : '<regexp>',
                 'help' : 'Regular expression which should only match correct '
                          'class names'}
                ),
               ('function-rgx',
                {'default' : FUNC_NAME_RGX,
                 'type' :'regexp', 'metavar' : '<regexp>',
                 'help' : 'Regular expression which should only match correct '
                          'function names'}
                ),
               ('argument-rgx',
                {'default' : VAR_NAME_RGX,
                 'type' :'regexp', 'metavar' : '<regexp>',
                 'help' : 'Regular expression which should only match correct '
                          'argument names'}),
               ('variable-rgx',
                {'default' : VAR_NAME_RGX,
                 'type' :'regexp', 'metavar' : '<regexp>',
                 'help' : 'Regular expression which should only match correct '
                          'variable names'}
                ),
               ('good-variable-names',
                {'default' : ('i', 'j', 'k', 'ex'),
                 'type' :'csv', 'metavar' : '<names>',
                 'help' : 'Good variable names which are lesser than \
min-name-length or wouldn\'t match the variable-rgx, separated by a comma'}
                ),
               ('bad-variable-names',
                {'default' : ('foo', 'bar', 'baz', 'toto', 'tutu', 'tata'),
                 'type' :'csv', 'metavar' : '<names>',
                 'help' : 'Bad variable names which are greater than \
min-name-length or would match the variable-rgx, separated by a comma'}
                ),
               )

    def open(self):
        self.linter.add_stats(module=0, constant=0, function=0, class_=0,
                              undocumented_module=0, undocumented_function=0,
                              undocumented_class=0, badname_module=0,
                              badname_class=0, badname_function=0,
                              badname_constant=0, badname_variable=0,
                              badname_argument=0)
        self.stats = self.linter.stats
        self._layouts = []
        self._returns = []
        self._branchs = []

    def visit_module(self, node):
        self.stats['module'] += 1
        self.check_name('module', node.__name__.split('.')[-1], node)
        self.check_docstring('module', node)
        self.check_required_attributes(node, self.config.required_attributes)
            
    def visit_classdef(self, node):
        self.stats['class'] += 1
        self.check_name('class', node.__name__, node)
        self.check_docstring('class', node)
        self.inc_branch()
        # check for redefinition
        self.check_redefinition('class', node)
            
    def visit_funcdef(self, node, close=0):
        is_method = node.is_method()
        if close:
            returns = self._returns.pop()
            if is_method and node.__name__ == '__init__' and returns:
                msg = '__init__ return'
                self.add_message('error', msg, node=node)
            elif returns > self.config.max_returns:
                msg = 'Too many return statements (%s/%s)'
                msg %= (returns, self.config.max_returns)
                self.add_message('warning', msg, node=node)
            branchs = self._branchs.pop()
            if branchs > self.config.max_branchs:
                msg = 'Too many branchs (%s/%s)'
                msg %= (branchs, self.config.max_branchs)
                self.add_message('warning', msg, node=node)
            return
        # init stats and counters for branchs and returns
        self.stats['function'] += 1
        self.inc_branch()
        self._returns.append(0)
        self._branchs.append(0)
        # function name
        self.check_name('function', node.__name__, node)
        # docstring
        if self.config.no_docstring_rgx.match(node.__name__) is None:
            self.check_docstring('function', node)
        # check default arguments'value
        if node.func_defaults:
            self.check_defaults(node.func_defaults, node)
        # check number of arguments
        args = node.args
        if len(args) > self.config.max_args:
            msg ='Too many arguments (%s/%s)'
            msg %= (len(args), self.config.max_args)
            self.add_message('warning', msg, node=node)
        # check arguments name
        self.recursive_check_names(args, node)
        # check number of local variables
        locnum = node.co_nlocals
        if locnum > self.config.max_locals:
            msg = 'Too many local variables (%s/%s)'
            msg %= (locnum, self.config.max_locals)
            self.add_message('warning', msg, node=node)
        # check local variable, avoiding argument, imported names, global names
        # and current class name if the function is actually a method
        for var, stmt in node.locals().items():
            if (not in_nested_list(args, var)
                and not isinstance(stmt, ImportStmt) 
                and not isinstance(stmt, GlobalStmt) 
                and not (is_method and var == node.parent.__name__)):
                self.check_name('variable', var, node)
        # check number of statements
        stmts_num = node.last_statement().num - node.num
        if stmts_num > self.config.max_statements:
            msg = 'Too many statements (%s/%s)'
            msg %= (stmts_num, self.config.max_statements)
            self.add_message('warning', msg, node=node)
        # check for redefinition
        self.check_redefinition(is_method and 'method' or 'function', node)

    def visit_returnstmt(self, node):
        self._returns[-1] += 1
        self.check_unreachable(node)
        
    def visit_yieldstmt(self, node):
        self._returns[-1] += 1
        self.check_unreachable(node)

    def visit_continuestmt(self, node):
        self.check_unreachable(node)

    def visit_breakstmt(self, node):
        self.check_unreachable(node)

    def visit_raisestmt(self, node):
        self.check_unreachable(node)

    def visit_ifstmt(self, node):
        self.inc_branch()
        
    def visit_elifstmt(self, node):
        self.inc_branch()
        
    def visit_elsestmt(self, node):
        self.inc_branch()
        
    def visit_whilestmt(self, node):
        self.inc_branch()
        
    def visit_forstmt(self, node):
        self.inc_branch()
        
    def visit_trystmt(self, node):
        self.inc_branch()

    def visit_globalstmt(self, node):
        if self.config.global_warning:
            msg = 'use of the global statement'
            self.add_message('warning', msg, node=node)
            
    def inc_branch(self):
        """increments the branchs counter"""
        branchs = self._branchs
        for i in range(len(branchs)):
            branchs[i] += 1

    def check_unreachable(self, node):
        """check unreachable code"""
        if self.config.unreachable:
            unreach_stmt = next_sibling(node)
            if unreach_stmt is not None:
                self.add_message('warning', 'Unreachable code',
                                 node=unreach_stmt)
        
    def check_redefinition(self, type, node):
        """check for redefinition of a function / method / class name"""
        if not self.config.redefinition:
            return
        defined_self = node.parent.locals()[node.__name__]
        if defined_self is not node:
            msg = '%s yet defined line %s' % (
                type, defined_self.line)
            self.add_message('error', msg, node=node)
        
    def check_docstring(self, type, node):
        """check the node has a non empty docstring 
        """
        docstring = node.docstring()
        if docstring is None:
            self.stats['undocumented_'+type] += 1
            self.add_message('warning', 'Missing docstring', node=node)
        elif not docstring.strip():
            self.stats['undocumented_'+type] += 1
            self.add_message('warning', 'Empty docstring', node=node)
            
    def recursive_check_names(self, args, node):
        """check names in a possibly recursive list <arg>
        """
        for arg in args:
            if type(arg) is type([]):
                self.recursive_check_names(arg, node)
            else:
                self.check_name('argument', arg, node)
    
    def check_name(self, type, name, node):
        """check for a name using the type's regexp"""
        regexp = getattr(self.config, type + '_rgx')
        if type in ('argument', 'variable'):
            if name in self.config.good_variable_names:
                return
            if name in self.config.bad_variable_names:
                msg = 'Invalid %s name "%s"'
                self.add_message('warning', msg % (type, name), node=node)
                self.stats['badname_' + type] += 1
                return
        if regexp.match(name) is None:
            msg = 'Invalid %s name "%s" (should match %s)'
            self.add_message('warning', msg % (type, name, regexp.pattern),
                             node=node)
            self.stats['badname_' + type] += 1
        elif len(name) < self.config.min_name_length:
            msg = 'Too short %s name "%s"'
            self.add_message('warning', msg % (type, name), node=node)
            self.stats['badname_' + type] += 1

    def check_defaults(self, defaults, node):
        """check for dangerous default values as arguments"""
        for default_value in defaults:
            if type(default_value) in (type([]), type({})):
                msg = 'Dangerous default value %r as argument'
                self.add_message('warning', msg % default_value, node=node)
        
    def check_required_attributes(self, node, attributes):
        """check for required attributes"""
        locals = node.locals()
        for attr in attributes:
            if not locals.has_key(attr):
                msg = 'Missing required attribute "%s"'
                self.add_message('warning', msg % attr, node=node)

    def make_stats(self, old_stats=None):
        """calculate some basic statistic :
    
        * percentage of different types documented
        * percentage of different types with a bad name
        """
        stats = self.stats
        nice_stats = {} 
        # percentage of different types documented and/or with a bad name
        sect = SectionLayout('Statistics by type')
        self._layouts.append(sect)
        for type in ('module', 'class', 'function'):
            nice_stats[type] = {}
            total = stats[type]
            if total == 0:
                doc_percent = 0
                badname_percent = 0
            else:
                undocumented = stats['undocumented_'+type]
                documented = total - undocumented
                doc_percent = float((documented)*100) / total
                badname = stats['badname_'+type]
                badname_percent = float((badname)*100) / total
            
            nice_stats[type]['percent_documented'] = doc_percent
            nice_stats[type]['percent_badname'] = badname_percent
        
        for type in ('constant', ):#'variable', 'argument'):
            nice_stats[type] = {}
            total = stats[type]
            if total == 0:
                badname_percent = 0
            else:
                badname = stats['badname_'+type]
                badname_percent = float((badname)*100) / total
            nice_stats[type]['percent_badname'] = badname_percent
        lines = [('type', 'number', 'old number', 'difference',
                  '%documented', '%badname')]
        for type in ('module', 'function', 'class'):
            new = stats[type]
            old = old_stats.get(type, None)
            if old is not None:
                diff_str = diff_string(old, new)
            else:
                old, diff_str = 'NC', 'NC'
            lines.append((type, str(new), str(old), diff_str,
                          '%.2f' % nice_stats[type]['percent_documented'],
                          '%.2f' % nice_stats[type]['percent_badname']))

        sect.append(TableLayout(lines, row_headers=1))

def next_sibling(node):
    """return the next sibling for the given node if any
    """
    parent = node.parent
    if parent is None:
        return None
    i = parent.children.index(node)
    try:
        return parent.children[i+1]
    except IndexError:
        return None
    
def register(linter):
    """required method to auto register this checker"""
    linter.register_checker(BasicChecker(linter))

