# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Copyright (c) 2000-2003 LOGILAB S.A. (Paris, FRANCE).
 http://www.logilab.fr/ -- mailto:contact@logilab.fr

 basic checker for Python code
"""

__revision__ = "$Id: classes.py,v 1.16 2003/10/03 06:03:44 syt Exp $"

import inspect    
from logilab.common.astng import FuncDef
from logilab.common.astutils import get_names
from logilab.pylint.interfaces import IASTNGChecker
from logilab.pylint.checkers import BaseChecker, ASTNGCheckerHandler


class ClassChecker(BaseChecker, ASTNGCheckerHandler):
    """checks for :                                                            
    * methods without self as first argument                                   
    * overriden methods signature                                              
    * access only to existant members via self                                 
    * attributes not defined in the __init__ method                            
    * supported interfaces implementation                                      
    * unreachable code                                                         
    """
    
    __implements__ = (IASTNGChecker,)# IStatable)

    name = 'classes'
    priority = -2
    options = (('self-warning',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help' : 'print a warning method without self as first \
argument'}
                ),
               ('init',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help': 'check that attributes are defined in the \
__init__ method'}               
                ),
               ('members',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help': 'check that accessed members are defined'}
                ),               
               ('interfaces',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help': 'check declared interfaces are truly implemented'}
                ),               
               )

    def visit_classdef(self, class_node, close=0):
        if close:
            # checks attributes are defined in __init__
            if self.config.init:
                for attr, node in class_node.instance_attrs.items():
                    frame = node.get_frame()
                    if frame.__name__ not in ('__init__', '__new__', 'setUp'):
                        parent_node = class_node.get_ancestor_for_attribute(attr)
                        if parent_node:
                            frame = parent_node.instance_attrs[attr].get_frame()
                            if frame.__name__ not in ('__init__', '__new__', 'setUp'):
                                msg = 'Attribute %s defined outside __init__'
                                self.add_message('warning', msg % attr, node=node)
                        else:
                            msg = 'Attribute %s defined outside __init__'
                            self.add_message('warning', msg % attr, node=node)
            # check access to existant members
            if self.config.members:
                for attr, nodes in self._accessed.items():
                    # is it a builtin attribute
                    if attr in ('__dict__', '__class__'):
                        continue
                    # is it an instance attribute ?
                    if class_node.instance_attrs.has_key(attr):
                        continue
                    # or a class attribute ?
                    if class_node.locals().has_key(attr):
                        continue
                    # or an inherited method / attribute ?
                    if class_node.get_ancestor_for_method(attr):
                        continue
                    if class_node.get_ancestor_for_attribute(attr):
                        continue
                    msg = 'Access to undefined member %s' % attr
                    for node in nodes:
                        self.add_message('error', msg, node=node)
                
            del self._accessed
        else:
            self._accessed = {}
            if self.config.interfaces:
                if class_node.object:
                    self.check_interfaces(class_node)

            
    def visit_funcdef(self, node, close=0):
        # check first argument is self if this is actually a method
        if node.is_method():
            if close:
                try:
                    del self._first_attr
                except:
                    pass
                return
            self._first_attr = node.args and node.args[0]
            if self.config.self_warning:
                # static method
                if node.is_static_method():
                    if node.args and node.args[0] in ('self', 'cls'):
                        msg = 'Static method with "%s" as first argument'
                        self.add_message('warning', msg % node.args[0],
                                         node=node)
                    self._first_attr = None
                # class / regular method with no args
                elif not node.args:
                    self.add_message('error', 'Method has no argument',
                                     node=node)
                # class method
                elif node.is_class_method():
                    if self._first_attr != 'cls':
                        msg = 'Class method doesn\'t have "cls" as first argument'
                        self.add_message('error', msg, node=node)
                # regular method without self as argument
                elif self._first_attr != 'self':
                    msg = 'Method doesn\'t have "self" as first argument'
                    self.add_message('error', msg, node=node)
                
            if node.__name__ == '__init__':
                return
            
            # check signature if the method overrload an herited method
            overriden = node.parent.get_ancestor_for_method(node.__name__)
            if overriden is not None:
                # get astng for the searched method
                try:
                    astng = overriden.locals()[node.__name__]
                    self.check_signature(node, astng, 'overriden')
                except KeyError:
                    # we have found the method but it's not in the local
                    # dictionnary.
                    # This may happen with astng build from living objects only
                    pass
            # check if the method overload an attribute
            overriden = node.parent.get_ancestor_for_attribute(node.__name__)
            if overriden:
                msg = 'Method hide an inherited attribute from %s'
                self.add_message('error', msg % overriden.__name__, node=node)
    
            
    def visit_exceptstmt(self, node):
        if node.ast is not None:
            for name in get_names(node.ast):
                self.handle_name(name, node)
        if node.instance_ast is not None:
            for name in get_names(node.instance_ast):
                self.handle_name(name, node)
            
    def visit_ifstmt(self, node):
        if node.ast:
            for name in get_names(node.ast):
                self.handle_name(name, node)
                
    visit_elifstmt = visit_whilestmt = visit_ifstmt
    
    def visit_printstmt(self, node):
        for ast in node.asts:
            for name in get_names(ast):
                self.handle_name(name, node)
                
    visit_returnstmt = visit_yieldstmt = visit_raisestmt = visit_printstmt

    def visit_expressionstmt(self, node):
        for name in get_names(node.ast):
            self.handle_name(name, node)
            
    def visit_forstmt(self, node):
        for name in get_names(node.in_ast):
            self.handle_name(name, node)
            
    def visit_assertstmt(self, node):
        for name in get_names(node.ast):
            self.handle_name(name, node)
        if node.msg_ast:
            for name in get_names(node.msg_ast):
                self.handle_name(name, node)
                
    def visit_execstmt(self, node):
        for name in get_names(node.ast):
            self.handle_name(name, node)
        if node.in_ast is not None:
            for name in get_names(node.in_ast):
                self.handle_name(name, node)
                
    def visit_assignmentstmt(self, node):
        """ get accessed members from the rhs """
        for ast in node.variables_asts:
            for name in get_names(ast):
                self.handle_name(name, node)
        for name in get_names(node.rhs_ast):
            self.handle_name(name, node)


    def handle_name(self, name, node):
        """ check if the name handle an access to a class member """
        if hasattr(self, '_first_attr'):
            parts = name.split('.')
            if parts[0] == self._first_attr and len(parts) > 1:
                self._accessed.setdefault(parts[1], []).append(node)

    def check_interfaces(self, class_node):
        """ check that the given class node really implements declared
        interfaces
        """
        try:
            implements = class_node.object.__implements__
        except AttributeError:
            return
        manager = self.linter.manager
        if not type(implements) in (type(()), type([])):
            implements = (implements,)
        for iface in implements:
            if not inspect.isclass(iface):
                msg = 'Interface %s is not a class (%s)' % (iface, type(iface))
                self.add_message('error', msg, node=class_node)
                continue
                
            iface_node = manager.astng_from_class(iface)
            iface_name = iface.__name__
            for member in dir(iface):
                member = getattr(iface, member)
                if not inspect.ismethod(member):
                    continue
                name = member.__name__
                # get class method astng
                try:
                    method_astng = class_node.get_method(name)
                except:
                    msg = 'unable to check method %s of interface %s'
                    self.add_message('error', msg % (name, iface_name),
                                     node=class_node)
                    import traceback
                    traceback.print_exc()
                    continue
                if method_astng is None:
                    msg = 'missing method %s from interface %s' 
                    self.add_message('error', msg % (name, iface_name),
                                     node=class_node)
                    continue
                if not method_astng.parent is class_node:
                    # don't go further if it is an inherited method
                    continue
                # get iface method astng
                try:
                    imethod_astng = iface_node.get_method(name)
                except:
                    msg = 'unable to check method %s of interface %s'
                    self.add_message('error', msg % (name, iface_name),
                                     node=class_node)
                    import traceback
                    traceback.print_exc()
                    continue
                if imethod_astng is None:
                    msg = 'Unable to get member %s from interface %s' 
                    self.add_message('warning', msg % (name, iface_name),
                                     node=class_node)
                    continue
                # check signature
                self.check_signature(method_astng, imethod_astng,
                                     '%s interface' % iface_name)
        
    
    def check_signature(self, method1, method2, type):
        """ check that the signature of the two given methods match """
        if not isinstance(method1, FuncDef):
            import sys
            print >>sys.stderr, '%s is not a FuncDef statement' % method1
            return
        if not isinstance(method2, FuncDef):
            import sys
            print >>sys.stderr, '%s is not a FuncDef statement' % method2
            return
        if len(method1.args) != len(method2.args):
            msg = 'Arguments number differs from %s method'
            self.add_message('warning', msg % type, node=method1)
        elif len(method1.func_defaults) != len(method2.func_defaults):
            msg = 'Signature differs from %s method'
            self.add_message('warning', msg % type, node=method1)
            
        
def register(linter):
    """ required method to auto register this checker """
    linter.register_checker(ClassChecker(linter))
