# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Copyright (c) 2000-2003 LOGILAB S.A. (Paris, FRANCE).
 http://www.logilab.fr/ -- mailto:contact@logilab.fr

Python code format's checker.

By default try to follow Guido's style guide :

http://www.python.org/doc/essays/styleguide.html

Some parts of the process_token method is based from The Tab Nanny std module.
"""

__revision__ = "$Id: format.py,v 1.18 2003/10/09 14:50:52 alf Exp $"

from __future__ import generators
from os import linesep
import sys
import re
import tokenize
if not hasattr(tokenize, 'NL'):
    raise ValueError("tokenize.NL doesn't exist -- tokenize module too old")


from logilab.pylint.interfaces import IRawChecker
from logilab.pylint.checkers import BaseRawChecker, CheckerHandler

STRING_RGX = re.compile(r"'([^\'\\]|\\.)*['\\]|\"([^\\\"\\]|\\.)*[\"\\]")

OPERATORS = r'!=|<=|==|>=|<|>|=|\+=|-=|\*=|/=|%'

OP_RGX_MATCH_1 = r'[^(]*(?<!\s|\^|<|>|=|\+|-|\*|/|!|%%|&|\|)(%s).*' % OPERATORS
OP_RGX_SEARCH_1 = r'(?<!\s|\^|<|>|=|\+|-|\*|/|!|%%|&|\|)(%s)' % OPERATORS

OP_RGX_MATCH_2 = r'[^(]*(%s)(?!\s|=|>|<).*' % OPERATORS
OP_RGX_SEARCH_2 = r'(%s)(?!\s|=|>)' % OPERATORS

BAD_CONSTRUCT_RGXS = (
    (re.compile(r'\s*(class|def|if|for|while)\s+([^:\[\]]|\[.*:?.*\])*?:\s*\w+(\s|\w)*'),
     re.compile(r':\s*[^\s]+.*'),
     'Colon not followed by a new line'),
    
    (re.compile(OP_RGX_MATCH_1),
     re.compile(OP_RGX_SEARCH_1),
     'Operator not preceded by a space'),
    
    (re.compile(OP_RGX_MATCH_2),
     re.compile(OP_RGX_SEARCH_2),
     'Operator not followed by a space'),
    
    (re.compile(r'.*,[^\s)].*'),
     re.compile(r',[^\s)]'),
     'Comma not followed by a space'),
    )


def searchall(rgx, line):
    """apply a regexp using "search" until no more match is found
    """
    match = rgx.search(line)
    while match is not None:
        yield match
        match = rgx.search(line, match.end())        
    
def get_string_coords(line):
    """return a list of string positions (tuple (start, end)) in the line
    """
    result = []
    for match in searchall(STRING_RGX, line):
        result.append( (match.start(), match.end()) )
    return result

def in_coords(match, string_coords):
    """return true if the match in in the string coord
    """
    mstart = match.start()
    #mend = match.end()
    for start, end in string_coords:
        if mstart >= start and mstart < end:
            return 1
    return 0

def check_line(line):
    """check a line for a bad construction
    if it founds one, return a message describing the problem
    else return None
    """
    no_strings = STRING_RGX.sub('', line)
    for rgx_match, rgx_search, msg in BAD_CONSTRUCT_RGXS:
        if rgx_match.match(no_strings):
            string_positions = get_string_coords(line)
            for match in searchall(rgx_search, line):
                if not in_coords(match, string_positions):
                    if line.endswith(linesep):
                        line = line[:-len(linesep)]
                    return '%s\n%s' % (msg, pretty_match(match, line))
            print >>sys.stderr, 'Unable to match %r', line

            
class IndentError(Exception):
    """ expression raised on ambiguous indentation """

 
class FormatChecker(BaseRawChecker, CheckerHandler):
    """checks for :                                                            
    * unauthorized constructions                                               
    * strict indentation                                                       
    * line length                                                             
    """
    
    __implements__ = IRawChecker

    # configuration section name
    name = 'format'
    # configuration options
    # for available dict keys/values see the optik parser 'add_option' method
    options = (('max-line-length',
                {'default' : 80, 'type' : "int", 'metavar' : '<int>',
                 'help' : 'Maximum number of characters on a single line.'}),
               ('max-module-lines',
                {'default' : 1000, 'type' : 'int', 'metavar' : '<int>',
                 'help': 'Maximum number of lines in a module'}
                ),
               ('indent-string',
                {'default' : '    ', 'type' : "string", 'metavar' : '<string>',
                 'help' : 'String used as indentation unit. This is usually \
"    " (4 spaces) or "\\t" (1 tab).'}),
               ('bad-construction',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help' : 'Indicate whether to check or not for bad construction'
                 }),
#               ('bad-construction-rgx',
#                {'default' : BAD_CONSTRUCT_RGX,
#                 'type' : "regexp", 'metavar' : '<regexp>',
#                 'help' : 'Regular expression which should match bad \
#constructions.'}),
               )
    
    def process_tokens(self, tokens):
        """ process tokens and search for :
    
         _ non strict indentation (i.e. not always using the <indent> parameter as
           indent unit)
         _ too long lines (i.e. longer than <max_chars>)
         _ optionally bad construct (if given, bad_construct must be a compiled
           regular expression).
        """
        indent = tokenize.INDENT
        dedent = tokenize.DEDENT
        newline = tokenize.NEWLINE
        junk = tokenize.COMMENT, tokenize.NL
        indent_string = self.config.indent_string
        # add 1 for the trailing '\n'
        max_chars = self.config.max_line_length + len(linesep) 
        bad_construct = self.config.bad_construction
        indents = [0]
        check_equal = 0
        line_num = 0
        long_lines = {}
        for (type, token, start, end, line) in tokens:
            line_num = start[0]

            # check for line's length
            if end[1] > max_chars and not long_lines.has_key(line_num):
                long_lines[line_num] = 1
                msg = 'Too long line (%s/%s)'
                msg %= (end[1] - len(linesep), max_chars - len(linesep))
                self.add_message('warning', msg, line_num)

            if type == newline:
                # a program statement, or ENDMARKER, will eventually follow,
                # after some (possibly empty) run of tokens of the form
                #     (NL | COMMENT)* (INDENT | DEDENT+)?
                # If an INDENT appears, setting check_equal is wrong, and will
                # be undone when we see the INDENT.
                check_equal = 1

            elif type == indent:
                check_equal = 0
                level, msg = indent_level(token, indent_string, indents[-1]+1)
                if msg is not None:
                    self.add_message('warning', msg, line_num)
                    indents.append(indents[-1]+1)
                else:
                    indents.append(level)

            elif type == dedent:
                # there's nothing we need to check here!  what's important is
                # that when the run of DEDENTs ends, the indentation of the
                # program statement (or ENDMARKER) that triggered the run is
                # equal to what's left at the top of the indents stack
                check_equal = 1
                if len(indents) > 1:
                    del indents[-1]

            elif check_equal and type not in junk:
                # this is the first "real token" following a NEWLINE, so it
                # must be the first token of the next program statement, or an
                # ENDMARKER; the "line" argument exposes the leading whitespace
                # for this statement; in the case of ENDMARKER, line is an empty
                # string, so will properly match the empty string with which the
                # "indents" stack was seeded
                if bad_construct:
                    msg = check_line(line)
                    if msg:
                        self.add_message('warning', msg, line_num)
                check_equal = 0
                level, msg = indent_level(line, indent_string, indents[-1])
                if msg is not None:
                    self.add_message('warning', msg, line_num)
                    
        if line_num > self.config.max_module_lines:
            msg = 'Too many lines in module (%s)' % line_num
            self.add_message('warning', msg)


def indent_level(string, indent, expected):
    """return the indent level of the string
    
    raise IndentError on ambiguous indentation
    """
    level = 0
    unit_size = len(indent)
    while string[:unit_size] == indent:
        string = string[unit_size:]
        level += 1
    suppl = ''
    while string and string[0] in ' \t':
        if string[0] != indent[0]:
            msg = 'Mixed tab and spaces'
            return level, msg
        suppl += string[0]
        string = string [1:]
    if level != expected or suppl:
        if indent[0] == '\t':
            msg = "Bad indentation. Found %s tabs, expected %s" % (
                level * unit_size + len(suppl), expected * unit_size)
        else:
            msg = "Bad indentation. Found %r spaces, expected %s" % (
                level * unit_size + len(suppl), expected * unit_size)
    else:
        msg = None
    return level, msg


def pretty_match(match, string, underline_char='^'):
    '''return a string whith the match location underlined
    '''
    start = match.start()
    end = match.end()
    start_line_pos = string.rfind(linesep, 0, start)
    if start_line_pos == -1:
        start_line_pos = 0
        result = []
    else:
        result = [string[:start_line_pos]]
        start_line_pos += len(linesep)
    offset = start - start_line_pos
    underline = ' ' * offset + underline_char * (end - start)
    end_line_pos = string.find(linesep, end)
    if end_line_pos == -1:
        string = string[start_line_pos:]
        result.append(string)
        result.append(underline)
    else:
        end = string[end_line_pos + len(linesep):]
        string = string[start_line_pos:end_line_pos]
        result.append(string)
        result.append(underline)
        result.append(end)
    return linesep.join(result)

def register(linter):
    """ required method to auto register this checker """
    linter.register_checker(FormatChecker(linter))


