# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Copyright (c) 2000-2003 LOGILAB S.A. (Paris, FRANCE).
 http://www.logilab.fr/ -- mailto:contact@logilab.fr

imports checkers for Python code
"""

__revision__ = "$Id: imports.py,v 1.8 2003/10/03 17:40:09 syt Exp $"

from logilab.common.modutils import is_standard_module, is_relative, \
     get_module_part, STD_LIB_DIR

from logilab.pylint.interfaces import IASTNGChecker
from logilab.pylint.checkers import BaseChecker, ASTNGCheckerHandler
from logilab.pylint.layouts import SectionLayout, VerbatimLayout


# utilities to represents imports as tree ######################################

def make_tree_defs(mod_files_list):
    """ get a list of 2-uple (module, list_of_files_which_import_this_module),
    it will return a dictionnary to represent this as a tree
    """
    tree_defs = {}
    for mod, files in mod_files_list:
        prefixes = mod.split('.')
        d = (tree_defs, ())
        for prefix in prefixes:
            d = d[0].setdefault(prefix, [{}, []])
        d[1] += files
    return tree_defs

def repr_tree_defs(data, indent_str=None):
    """ return a string which represents imports as a tree"""
    s = ''
    l = data.items()
    for i in range(len(l)):
        mod, (sub, files) = l[i]
        if not files: files = ''
        if indent_str is None:
            s = '%s\n%s %s\n' % (s, mod, files)
            sub_indent_str = '  '
        else:
            s = '%s%s\-%s %s\n' % (s, indent_str, mod, files)
            if i == len(l)-1:
                sub_indent_str = '%s  ' % indent_str
            else: sub_indent_str = '%s| ' % indent_str
        s = '%s%s'% (s, repr_tree_defs(sub, sub_indent_str))
    return s


class ImportsChecker(BaseChecker, ASTNGCheckerHandler):
    """checks for                                                              
    * external modules dependancies                                            
    * relative / wildcard imports                                                         
    * cyclic imports                                                           
    * uses of depreceted modules
    """
    
    __implements__ = IASTNGChecker

    name = 'imports'
    priority = -1
    
    options = (('dependencies',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help': 'check external dependencies'}),
               ('relative',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help': 'check relative imports'}),               
               ('wildcard',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help': 'check wildcard imports'}),               
               ('cycles',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help': 'check cyclic imports'}),
               ('deprecated-modules',
                {'default' : ('regsub','TERMIOS'), 'type' : 'csv',
                 'metavar' : '<modules>',
                 'help' : 'Deprecated modules which should not be used, \
separated by a comma'}
                ),
               )

    def open(self):
        """ called before visiting project (i.e set of modules) """
        if self.config.dependencies:
            self.linter.add_stats(dependencies={})
        if self.config.cycles:
            self.linter.add_stats(cycles=[])
        self.stats = self.linter.stats
        self.import_graph = {}
        
    def close(self):
        """ called before visiting project (i.e set of modules) """
        if self.config.cycles:
            from logilab.common import get_cycles
            for cycle in get_cycles(self.import_graph):
                msg = 'cyclic import (%s)' % ' -> '.join(cycle)
                self.add_message('warning', msg)
        
    def visit_importstmt(self, node):
        """ triggered when an import statement is seen """            
        if node.base:
            self.imported_module(node.base, node)
        else:
            for name in node.names:
                self.imported_module(name, node)
        if self.config.wildcard and node.wildcard:
            self.add_message('warning', 'wildcard import', node=node)
        
    def imported_module(self, mod_path, node):
        """ notify an imported module, used to analyze dependancies
        """
        # check this isn't a deprecated module
        for mod_name in self.config.deprecated_modules:
            if mod_path[:len(mod_name)] == mod_name:
                msg = 'Uses of a deprecated module (%s)' % mod_path
                self.add_message('warning', msg, node)
        linter = self.linter
        context_file = linter.current_module.file
        relative = is_relative(mod_path, context_file)
        # check for relative import
        if self.config.relative and relative:
            msg = 'relative import (%s)'
            self.add_message('warning', msg % mod_path, node=node)
        # get module part
        try:
            imported = get_module_part(mod_path, context_file)
        except ValueError: # FIXME : necessary ?
            msg = 'Unable to import module %s from %s' 
            self.add_message('error', msg % (mod_path, context_file), node=node)
            import traceback
            traceback.print_exc()
            return
        except ImportError:
            msg = 'Unable to import module %s from %s'
            self.add_message('info', msg % (mod_path, context_file), node=node)
            return
        if relative:
            imported = '%s.%s' % (linter.current_module.__name__, imported)
        # check cycles
        if self.config.cycles:
            here = linter.current_module.__name__
            self.import_graph.setdefault(here, []).append(imported)
            self.import_graph.setdefault(imported, [])
        # handle dependancies
        try:
            if self.config.dependencies and not \
                   is_standard_module(mod_path, (STD_LIB_DIR, linter.base)):
                path = '.'.join([p[1] for p in linter.get_path()])
                list = self.stats['dependencies'].setdefault(mod_path, [])
                list.append(path)
        except:
            import sys
            print >>sys.stderr,mod_path
            import traceback
            traceback.print_exc()

    def get_layouts(self):
        layouts = []
        if self.config.dependencies:
            dep_info = make_tree_defs(self.stats['dependencies'].items())
            if dep_info:
                sect = SectionLayout('External dependencies')
                tree_str = repr_tree_defs(dep_info)
                sect.append(VerbatimLayout(tree_str))
            else:
                sect = SectionLayout('No external dependencies')
            layouts.append(sect)
        return layouts
    
            
def register(linter):
    """ required method to auto register this checker """
    linter.register_checker(ImportsChecker(linter))
