# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Copyright (c) 2000-2003 LOGILAB S.A. (Paris, FRANCE).
 http://www.logilab.fr/ -- mailto:contact@logilab.fr

Check source code is ascii only or has an encoding declaration (PEP 263)
"""

__revision__ = '$Id: misc.py,v 1.3 2003/10/04 15:33:09 syt Exp $'

import re

from logilab.pylint.interfaces import IRawChecker
from logilab.pylint.checkers import BaseChecker, CheckerHandler

def is_ascii(string):
    if string:
        return max(map(ord, string)) < 128
    return 0
    

EMACS_ENCODING_RGX = re.compile('[^#]*[#\s]*-\*-\s*coding: ([^\s]*)\s*-\*-\s*')
VIM_ENCODING_RGX = re.compile('[^#]*[#\s]*vim:fileencoding=\s*([^\s]*)\s*')
XML_ENCODING_RGX = re.compile('<\?xml version=[^\s]*\s*encoding=([^\s]*)\s*\?>')
CHARSET_RGX = re.compile('charset=([^\s"]*)')


def guess_encoding(buffer):
    """try to guess encoding from a buffer

    FIXME: it could be mime type driven but it seems less painful like that
    """
    assert type(buffer) is type(''), type(buffer)
    # default to ascii on empty buffer
    if not buffer:
        return 'ascii'
    
    # check for UTF-8 byte-order mark
    if buffer.startswith('\xef\xbb\xbf'):
        return 'UTF-8'
    
    first_lines = buffer.split('\n')[:2]
    for line in first_lines:
        # check for emacs encoding declaration
        m = EMACS_ENCODING_RGX.match(line)
        if m is not None:
            return m.group(1)
        # check for vim encoding declaration
        m = VIM_ENCODING_RGX.match(line)
        if m is not None:
            return m.group(1)
        
    # check for xml encoding declaration
    if first_lines[0].startswith('<?xml'):
        m = XML_ENCODING_RGX.match(first_lines[0])
        if m is not None:
            return m.group(1)[1:-1]
        # xml files with no encoding declaration default to UTF-8
        return 'UTF-8'
    
    # try to get charset declaration
    # FIXME: should we check it's html before ?
    m = CHARSET_RGX.search(buffer)
    if m is not None:
        return m.group(1)
    return None


class EncodingChecker(BaseChecker, CheckerHandler):
    """checks for:                                                             
    * source code with non ascii characters but np encoding declaration (PEP
      263)                                                                     
    * warning notes in the code like FIXME, XXX
    """
    __implements__ = IRawChecker

    # configuration section name
    name = 'miscellaneous'

    _msg1 = 'Non ascii characters found but no encoding specified (PEP 263)'
    _msg2 = 'Wrong encoding specified (%s)'
    _msg3 = 'Unknown encoding specified (%s)'

    options = (('encoding',
                {'type' : 'yn', 'metavar' : '<y_or_n>',
                 'default' : 1, 
                 'help' : 'check or not for source code encoding'
                 }),
               
               ('notes',
                {'type' : 'csv', 'metavar' : '<comma separated values>',
                 'default' : ('FIXME', 'XXX'),
                 'help' : 'list of notes to take in consideration, separated \
by a comma. Default to FIXME,XXX'
                 }),
               
               )
    def process_module(self, module_name, stream):
        # source encoding
        data = stream.read()
        if not is_ascii(data):
            encoding = guess_encoding(data)
            if encoding is None:
                self.add_message('error', self._msg1, 1)
            else:
                try:
                    unicode(data, encoding)
                except UnicodeError:
                    self.add_message('error', self._msg2 % encoding, 1)
                except LookupError:
                    self.add_message('error', self._msg3 % encoding, 1)
        del data
        # warning notes in the code
        if not self.config.notes:
            return
        stream.seek(0)
        notes = []
        for note in self.config.notes:
            notes.append(re.compile(note))
        linenum = 1
        for line in stream.readlines():
            for note in notes:
                match = note.search(line)
                if match:
                    self.add_message('warning', line[match.start():-1], linenum)
                    break
            linenum += 1
                    
                
            
def register(linter):
    """ required method to auto register this checker """
    linter.register_checker(EncodingChecker(linter))
