# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Copyright (c) 2000-2003 LOGILAB S.A. (Paris, FRANCE).
 http://www.logilab.fr/ -- mailto:contact@logilab.fr

variables checkers for Python code
"""

__revision__ = "$Id: variables.py,v 1.19 2003/10/04 09:57:54 syt Exp $"

from copy import copy
from logilab.common.astng import ImportStmt, StringStmt, RaiseStmt, PassStmt, \
     GlobalStmt
from logilab.common.astutils import get_names, get_lhs_consumed_names, \
     get_rhs_consumed_names
from logilab.pylint.interfaces import IASTNGChecker
from logilab.pylint.checkers import BaseChecker, ASTNGCheckerHandler


def is_interface(classe):
    for base in classe.__bases__:
        if base.__name__ == 'Interface':
            return 1
        elif is_interface(base):
            return 1
    return 0


def is_builtin(name):
    """ return true if <name> could be considered as a builtin """
    try:
        eval(name, {}, {})
        return 1
    except NameError:
        if name == '__path__':
            return 1
        return 0

def is_abstract(func_def):
    result = 0
    for c in func_def:
        if isinstance(c, StringStmt):
            continue
        elif (isinstance(c, RaiseStmt) and 
              get_names(c.ast) == ['NotImplementedError']):
            result = 1
        elif isinstance(c, PassStmt):
            result = 1
        else:
            result = 0
    return result


class VariablesChecker(BaseChecker, ASTNGCheckerHandler):
    """checks for                                                              
    * unused variables / imports                                               
    * undefined variables                                                      
    * redefinition of variable from builtins or from an outer scope            
    * use of variable before assigment                                         
    """
    
    __implements__ = IASTNGChecker

    name = 'variables'
    priority = -1
    
    options = ()

    def visit_module(self, node, close=0):
        """ check globals
        """
        if not close:
            self._to_consume = [(copy(node.locals()), {})]
        else:
            assert len(self._to_consume) == 1
            not_consumed = self._to_consume.pop()[0]
            for name, stmt in not_consumed.items():
                if isinstance(stmt, ImportStmt) and stmt.base != '__future__':
                    msg = 'Unused import %s' % name
                    self.add_message('warning', msg, node=stmt)
            del self._to_consume

    def visit_classdef(self, node, close=0):
        """ check class' locals
        """
        if not close:
            self._to_consume.append((copy(node.locals()), {}))
            for base in node.bases:
                self.check_name(base.split('.')[0], node)
        else:
            # do not check for not used locals here (no sense)
            self._to_consume.pop()
            
    def visit_funcdef(self, node, close=0):
        """ check function's locals
        """
        if not close:
            globals = node.globals()
            for name, stmt in node.locals().items():
                if globals.has_key(name) and not isinstance(stmt, GlobalStmt):
                    msg = 'Redefining variable (%s) from outer scope (line %s)'
                    line = globals[name].line
                    self.add_message('warning', msg % (name, line), node=stmt)
                elif is_builtin(name):
                    msg = 'Redefining built-in (%s)' % name
                    self.add_message('warning', msg, node=stmt)
                    
            self._to_consume.append((copy(node.locals()), {}))
            for ast in node.defaults_asts:
                for name in get_rhs_consumed_names(ast):
                    self.check_name(name, node)
        else:
            not_consumed = self._to_consume.pop()[0]
            if node.is_method() and (is_interface(node.parent) or is_abstract(node)):
                return
            for name, stmt in not_consumed.items():
                if isinstance(stmt, GlobalStmt):
                    continue
                msg = 'Unused variable %s' % name
                self.add_message('warning', msg, node=stmt)

        
    def visit_ifstmt(self, node):
        if node.ast:
            for name in get_rhs_consumed_names(node.ast):
                self.check_name(name, node)
    
    visit_elifstmt = visit_whilestmt = visit_ifstmt
    
    def visit_raisestmt(self, node):
        for ast in node.asts:
            for name in get_rhs_consumed_names(ast):
                self.check_name(name, node)
    
    visit_printstmt = visit_returnstmt = visit_yieldstmt = visit_raisestmt
            
    def visit_exceptstmt(self, node):
        if node.ast is None:
            return
        for name in get_rhs_consumed_names(node.ast):
            self.check_name(name, node)
        if node.instance_ast is None:
            return
        for name in get_rhs_consumed_names(node.instance_ast):
            self.check_name(name, node)
                

    def visit_assertstmt(self, node):
        for name in get_rhs_consumed_names(node.ast):
            self.check_name(name, node)
        if node.msg_ast is None:
            return
        for name in get_rhs_consumed_names(node.msg_ast):
            self.check_name(name, node)
        
    def visit_execstmt(self, node):
        for name in get_rhs_consumed_names(node.ast):
            self.check_name(name, node)
        if node.in_ast is not None:
            for name in get_rhs_consumed_names(node.in_ast):
                self.check_name(name, node)
                
    def visit_forstmt(self, node):
        for name in get_rhs_consumed_names(node.in_ast):
            self.check_name(name, node)

    def visit_assignmentstmt(self, node):
        for name in get_rhs_consumed_names(node.rhs_ast):
            self.check_name(name, node)
        for ast in node.variables_asts:
            for name in get_lhs_consumed_names(ast):
                self.check_name(name, node)
            
    def visit_expressionstmt(self, node):
        for name in get_rhs_consumed_names(node.ast):
            self.check_name(name, node)

    def visit_globalstmt(self, node):
        globs = node.globals()
        # FIXME: backward compat with astng, remove latter
        if not hasattr(node, 'variables'):
            variables = []
            for ast in node.asts:
                variables += get_names(ast)
        else:
            variables = node.variables
        # end backward compat
        for name in variables:
            if not globs.has_key(name):
                msg = 'Undefined global %s' % name
                self.add_message('warning', msg, node=node)
                
    def check_name(self, name, stmt):
        """ check that a name is defined if the current scope
        """
        for i in range(len(self._to_consume) - 1, -1, -1):
            to_consume, consumed = self._to_consume[i]
            if consumed.has_key(name):
                break
            try:
                consumed[name] = def_stmt = to_consume[name]
                del to_consume[name]
                if def_stmt and stmt.locals() is consumed[name].locals() and \
                       stmt.line < def_stmt.line:
                    msg = 'Using variable (%s) before assigment' % name
                    self.add_message('error', msg, node=stmt)
                break
            except KeyError:
                continue
        else:
            if not is_builtin(name):
                self._to_consume[-1][1][name] = 1
                msg = 'Undefined variable (%s)'
                self.add_message('error', msg % name, node=stmt)

    
def register(linter):
    """ required method to auto register this checker """
    linter.register_checker(VariablesChecker(linter))
