# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Copyright (c) 2002-2003 LOGILAB S.A. (Paris, FRANCE).
 http://www.logilab.fr/ -- mailto:contact@logilab.fr

Interfaces for PyLint objects
"""

__revision__ = "$Id: interfaces.py,v 1.3 2003/06/19 07:15:36 alf Exp $"

from logilab.common.interface import Interface
from logilab.common.astng import IASTNGVisitor


class IChecker(Interface):
    """ This is an base interface, not designed to be used elsewhere than for
    sub interfaces definition.
    """

    def set_option(self, option, value):
        """ method called to set an option
        """
        
    def open(self):
        """ called before visiting project (i.e set of modules) """
        
    def close(self):
        """ called after visiting project (i.e set of modules) """

    def open_module(self):
        """ called before visiting a module """
        
    def close_module(self):
        """ called after visiting a module """

    def get_layouts(self):
        """ return a list of layouts objects to control results display """

        
class IStatable(Interface):
    """add-on interface for checkers which are able to make stats on their data
    """

    def make_stats(self, old_stats):
        """ called after all visit to calculate statistics on collected data

        optional old_stats is a stats dictionnary from a previous run
        (maybe an empty dict)
        """
        
    
class IRawChecker(IChecker):
    """ interface for checker which need to parse the raw file
    """
    
    def process_module(self, name, stream):
        """ process the module with the given name
        
        the module's content is accessible via the stream object
        """


class IASTNGChecker(IChecker, IASTNGVisitor):
    """ interface for checker which prefers receive events according to
    statement type
    """


class ILinter(Interface):
    """ interface for the linter class

    the linter class will generate events to its registered checkers.
    Each ckecker may interact with the linter instance using this API
    """
    
    def register_checker(self, checker):
        """ register a new checker class

        checker is a class implementing IrawChecker or / and IASTNGChecker
        """

    def add_message(self, type, msg, line=None):
        """ add a message of a given type

        type must be in ('info', 'warning', 'error', 'fatal')
        raw checkers should provide the line argument
        """
        

class IReporter(Interface):
    """ reporter collect messages and display results encapsulated in a layout
    """
    def add_message(self, type, path, line, msg):
        """ add a message of a given type

        type maybe in ('info', 'warning', 'error', 'fatal')
        path represents the current location in code
        line is the current line number
        """
        
    def display_results(self, layout):
        """ display results encapsulated in the layout tree
        """
        
    
__all__ = ('IRawChecker', 'IStatable', 'IASTNGChecker', 'ILinter', 'IReporter')
