#!/usr/bin/env python
# Copyright (c) 2002-2003 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" %prog [options] module_or_package

  Check that a module satisfy a coding standard
"""

__revision__ = "$Id: lint.py,v 1.30 2003/10/10 09:14:12 syt Exp $"

from __future__ import nested_scopes

import sys
import os
from os.path import dirname

from logilab.common.configuration import OptionsManagerMixIn
from logilab.common.astng import ASTNG_MANAGER, ASTNGBuildingException, Module
from logilab.common.modutils import modpath_from_file, get_modules
from logilab.common.interface import implements

from logilab.pylint.interfaces import ILinter, IRawChecker, IASTNGChecker, \
     IStatable
from logilab.pylint.checkers import ASTNGCheckerHandler, BaseChecker
from logilab.pylint import config
from logilab.pylint.layouts import SectionLayout, TableLayout
from logilab.pylint.reporters import diff_string, split_path
from logilab.pylint.reporters.text import TextReporter

from logilab.pylint.__pkginfo__ import version

# utilities ###################################################################

def get_note_message(note):
    """return a message according to note
    note is a float < 10  (10 is the highest note)
    """
    assert note <= 10, "Note is %.2f. Either you cheated, or pylint's broken!"%note
    if note < 0:
        return 'You have to do something quick !'
    elif note < 1:
        return 'Hey! This is really dreadful. Or maybe pylint is buggy?'
    elif note < 2:
        return "Come on! You can't be proud of this code"
    elif note < 3:
        return 'Hum... Needs work.'
    elif note < 4:
        return 'Wouldn\'t you be a bit lazy?'
    elif note < 5:
        return 'A little more work would make it acceptable.'
    elif note < 6:
        return 'Just the bare minimum. Give it a bit more polish. '
    elif note < 7:
        return 'This is okay-ish, but I\'m sure you can do better.'
    if note < 8:
        return 'If you commit now, people should not be making nasty comments about you on c.l.py'
    elif note < 9:
        return 'That\'s pretty good. Good work mate.'
    elif note < 10:
        return 'So close to being perfect...'
    return 'Wow ! Now this deserves our uttermost respect.\nPlease send your code to python-projects@logilab.org'

def sort_checkers(checkers, enabled_only=1):
    """return a list of enabled checker sorted by priority
    """
    if enabled_only:
        checkers = [(-c.priority, c) for c in checkers
                    if c.is_enabled()]
    else:
        checkers = [(-c.priority, c) for c in checkers]
    checkers.sort()
    return [item[1] for item in checkers]


def get_module_and_frameid(node):
    """return the module name and the frame id in the module
    """
    frame = node.get_frame()
    module, object = '', []
    while frame:
        if isinstance(frame, Module):
            module = frame.__name__
        else:
            object.append(frame.__name__)
        try:
            frame = frame.parent.get_frame()
        except:
            frame = None
    object.reverse()
    return module, '.'.join(object)
        
# Python Linter class #########################################################

class PyLinter(OptionsManagerMixIn, BaseChecker, ASTNGCheckerHandler):
    """lint Python modules using external checkers.
    """

    __implements__ = (ILinter, IASTNGChecker, IStatable)
    
    name = 'master'
    priority = 0
    may_be_disabled = 0
    
    options = (("ignore",
                {'action' :"append", 'type' : "string", 'metavar' : "<file>",
                 'dest' : "black_list", "default" : ('CVS',),
                 'help' : """add <file> (may be a directory) to the black list\
. It should be a base name, not a path. You may set this option multiple times\
."""}),
               ("statistics",
                {'default': 1, 'type' : 'yn', 'metavar' : '<y_or_n>',
                 'help' : 'compute statistics on collected data.'}),
               ("persistent",
                {'default': 1, 'type' : 'yn', 'metavar' : '<y_or_n>',
                 'help' : 'pickle collected data for later comparisons.'}),
               ("evaluation",
                {'type' : 'string', 'metavar' : '<python_expression>',
'default': '10.0 - ((float(2 * errors + warnings) / statements) * 10)', 
                 'help' : 'Python expression which should return a note less \
than 10 (10 is the highest note).You have access to the variables errors, \
warnings, statements which respectivly contain the number of errors / warnings\
 messages and the total number of statements analyzed.'}),
               ("comment",
                {'default': 0, 'type' : 'yn', 'metavar' : '<y_or_n>',
                 'help' : 'Add a comment according to your evaluation note.'}),
               )
    
    def __init__(self, options=(), reporter=TextReporter()):
        self.options = options + PyLinter.options        
        OptionsManagerMixIn.__init__(self, usage=__doc__, version=version,
                                     config_file=config.PYLINTRC)
        BaseChecker.__init__(self)
        # helpers :
        self.manager = ASTNG_MANAGER
        self.reporter = reporter
        # base path of the current module / package analysed
        self.base = None
        # checkers
        self._checkers = {}
        self.register_checker(self)
        # visit variables
        self.current_module = None
        self.path = None
        self.base_name = None
        self.stats = None
        
    def add_stats(self, **kwargs):
        """add some stats entries to the statistic dictionary
        raise an AssertionError if there is a key conflict
        """
        for key, value in kwargs.items():
            if key[-1] == '_':
                key = key[:-1]
            assert not self.stats.has_key(key)
            self.stats[key] = value

        
    def add_message(self, type, msg, line=None, node=None):
        """add a message of a given type

        raw checkers should provide the line argument
        """
        type = type.lower()
        assert type in ('info', 'warning', 'error', 'fatal')
        key = type + 's'
        self.stats[key] += 1
        self.stats['by_module'][self.current_module.__name__][key] += 1
        if line is None and node is not None:
            line = node.line
        if node is None:
            module, object = split_path(self.path)
        else:
            module, object = get_module_and_frameid(node)
            
        self.reporter.add_message(type, module, object, line or 0, msg)
                        

    def get_path(self):
        """return a reference to the current path

        this method should only be used by astng checkers
        """
        return self.path


    def display_results(self):
        """assemble layouts and display them
        """
        # assemble layouts
        sect = SectionLayout('Report for %s' % self.base_name,
                             '%s statements analysed in %s.'% (
            self.stats['statements'], self.base or '?unknown location?'))
        for checker in sort_checkers(self._checkers.keys()):
            for layout in checker.get_layouts():
                sect.append(layout)
        if self.config.evaluation:
            # get a global note for the code
            note_sect = SectionLayout('Global evaluation')
            sect.append(note_sect)
            evaluation = self.config.evaluation
            try:
                note = eval(evaluation, {}, self.stats)
                msg = 'Your code has been rated at %.2f / 10.' % note
                if self.config.comment:
                    msg = '%s\n%s' % (msg, get_note_message(note))
            except Exception, e:
                msg = 'An exception occured while rating: %s' % e
            note_sect.description = msg
        self.reporter.display_results(sect)
        
        
    # checkers manipulation methods ###########################################
    
    def register_checker(self, checker):
        """register a new checker

        checker is an object implementing IRawChecker or / and IASTNGChecker
        """
        assert checker.priority <= 0, 'checker priority can\'t be >= 0'
        self._checkers[checker] = 1
        self.register_options_provider(checker)

    def disable_all_checkers(self):
        """disable all possible checkers """
        for checker in self._checkers.keys():
            checker.enable(0)


    # code checking methods ###################################################
    
    def check(self, mod_name):
        """main checking entry : check a module or package from its name
        if mod_name is a package, recurse on its subpackages / submodules
        """
        # get the given module representation
        self.base_name = mod_name
        try:
            astng = self.manager.astng_from_module_name(mod_name)
        except:
            print 'An error occured while parsing %s' % mod_name
            raise
        if astng.file is None:
            raise Exception('Unable to locate module %s\'s file' % mod_name)
        # set base
        self.base = base = dirname(astng.file)
        
        # notify global begin
        checkers = sort_checkers(self._checkers.keys())
        rev_checkers = checkers[:]
        rev_checkers.reverse()
        for checker in checkers:
            checker.open()
            
        # check this module
        self.check_astng_module(astng, checkers, rev_checkers)
        
        # recurse in package except if __init__ was explicitly given
        if not mod_name.endswith('.__init__') and astng.is_package():
            # recurse on others packages / modules if this is a package
            for mod_name in get_modules(mod_name, base,
                                        self.config.black_list):
                try:
                    # get the module representation
                    astng = self.manager.astng_from_module_name(mod_name)
                    # and check it
                    self.check_astng_module(astng, checkers, rev_checkers)
                except ASTNGBuildingException, e:
                    self.add_message('fatal', str(e), 0)
                except KeyboardInterrupt:
                    raise
                except Exception, e:
                    import traceback
                    traceback.print_exc()
                    self.add_message('fatal', '%s: %s' % (e.__class__, e), 0)
            
        # notify global end
        for checker in rev_checkers:
            checker.close()


    def check_astng_module(self, astng, checkers, rev_checkers):
        """check a module for its astng representation
        """
        # this is required to make works relative imports in analyzed code
        sys.path.insert(0, dirname(astng.file))
        try:
            self.current_module = astng
            mod_name = astng.__name__
            # notify visit begin
            for checker in checkers:
                checker.open_module()
            # call raw checkers if possible
            if not astng.is_pure_python():
                msg = 'Unable to run raw checkers on built-in module %s'
                self.add_message('info', msg % mod_name)
            else:
                try:
                    stream = open(astng.file)
                    for checker in checkers:
                        if implements(checker, IRawChecker):
                            stream.seek(0)
                            checker.process_module(mod_name, stream)
                except IOError, e:
                    msg = 'Unable to load file %s (%s)'
                    self.add_message('fatal', msg % (file, e))

            # generate events to astng checkers
            self.astng_events(astng, [c for c in checkers
                                      if implements(c, IASTNGChecker)])

            # notify visit end
            for checker in rev_checkers:
                checker.close_module()
        finally:
            sys.path.pop(0)
    
    def astng_events(self, astng, checkers):
        """generate event to astng checkers according to the current astng
        node and recurse on its children
        """
        self.stats['statements'] += 1
        # generate events for this node on each checkers
        for checker in checkers:
            astng.accept(checker)
        # recurse on children
        for child in astng.children:
            self.astng_events(child, checkers)
        checkers.reverse()
        for checker in checkers:
            try:
                astng.accept(checker, close=1)
            except TypeError:
                continue
        

    # IASTNGChecker interface #################################################
        
    def open(self):
        """initialize counters """
        self.stats = { 'statements': 0,
                       'infos'     : 0,
                       'warnings'  : 0,
                       'errors'    : 0,
                       'fatals'    : 0,
                       'by_module' : {},
                       }
    
    def close(self):
        """close the whole package /module, it's time to make stats
        if persistent run, pickle results for later comparison
        """
        # load old results if any
        old_stats = config.load_results(self.base_name)
        if self.config.statistics:
            # make statistics
            for checker in sort_checkers(self._checkers.keys()):
                if implements(checker, IStatable):
                    checker.make_stats(old_stats)
        # save results if persistent run
        if self.config.persistent:
            config.save_results(self.stats, self.base_name)
    
    def open_module(self):
        """initialize path and counters for a given module name"""
        name = self.current_module.__name__
        self.stats['by_module'][name] = {}
        self.stats['by_module'][name]['infos'] = 0
        self.stats['by_module'][name]['warnings'] = 0
        self.stats['by_module'][name]['errors'] = 0
        self.stats['by_module'][name]['fatals'] = 0
        self.path = [('module', name)]
        
    def close_module(self):
        """maintain path information """
        assert len(self.path) == 1, '%s %s' % (len(self.path), self.path)
        self.path.pop()
        
    def visit_classdef(self, node, close=0):
        """maintain path information """
        if close:
            self.path.pop()
        else:
            self.path.append(('class_', node.__name__))
    
    def visit_funcdef(self, node, close=0):
        """maintain path information """
        if close:
            self.path.pop()
        else:
            self.path.append(('function', node.__name__))
    
    def make_stats(self, old_stats):
        """called after all visit to calculate statistics on collected data

        optional old_stats is a stats dictionnary from a previous run
        """
        stats = self.stats
        
        sect = SectionLayout('Errors / warnings by module')
        self._layouts.append(sect)
        by_mod = {} 
        for m_type in ('fatals', 'errors', 'warnings'):
            total = stats[m_type]
            for module in stats['by_module'].keys():
                mod_total = stats['by_module'][module][m_type]
                if total == 0:
                    percent = 0
                else:
                    percent = float((mod_total)*100) / total
                by_mod.setdefault(module, {})[m_type] = percent
            
        sorted_result = []
        for module, mod_info in by_mod.items():
            sorted_result.append((mod_info['errors'],
                                  mod_info['warnings'],
                                  module))
        sorted_result.sort()
        sorted_result.reverse()
        lines = [('module', 'error', 'warning')]
        for line in sorted_result:
            lines.append((line[2], '%.2f' % line[0], '%.2f' % line[1]))
        sect.append(TableLayout(lines, row_headers = 1))
        
        sect = SectionLayout('Total errors / warnings')
        self._layouts.append(sect)
        lines = [('type', 'number', 'previous', 'difference')]
        for m_type in ('warnings', 'errors'):
            new = stats[m_type]
            old = old_stats.get(m_type, None)
            if old is not None:
                diff_str = diff_string(old, new)
            else:
                old, diff_str = 'NC', 'NC'
            lines.append((m_type, str(new), str(old), diff_str))
        sect.append(TableLayout(lines, row_headers = 1))
        

        
# this may help to import modules using gettext ###############################

try:
    __builtins__._ = str
except:
    __builtins__['_'] = str


# utilities ###################################################################

class Run:
    """helper class to use as main for pylint :
    
    run(*sys.argv[1:])
    """
    
    def __init__(self, *args):
        # export this variable since some packages may want to know they are
        # imported by pylint
        os.environ['PYLINT_IMPORT'] = '1'
        from logilab.pylint import checkers
        self.linter = linter = PyLinter((
            ("disable-all",
             {'action' : "callback",
              'callback' : self.disable_all_checkers,
              'help' : '''disable all possible checkers. This option should 
              precede enable-* options.'''}),
            ("generate-rcfile",
             {'action' : "callback", 'callback' : self.generate_config,
              'help' : '''generate a sample configuration file according to \
the current configuration. You can put other options before this one to use \
them in the configuration. This option causes the program to exit'''}),
            ("html",
            {'action' : "callback", 'callback' : self.set_format,
             'help' : "Use HTML as output format instead of text"}),
            ("zope",
            {'action' : "callback", 'callback' : self.init_zope,
             'help' : "Initialize Zope products before starting"}),
            
            ))
        # register checkers
        checkers.initialize(linter)
        # add some help section
        linter.add_help_section('Environment variables', config.ENV_HELP)
        linter.add_help_section('Output', '''
Using the default text output, the message format is :                         
        MESSAGE_TYPE: LINE_NUM:[OBJECT:] MESSAGE                               
There are 3 kind of message types :                                            
    * (W) warning and (E) error                                                
    these message types are used to distinguish the gravity of the detected    
problem.                                                                       
    * (F) fatal                                                                
    an error occured which prevented pylint from doing further processing.     
        ''')
        # load configuration
        linter.load_file_configuration()
        args = linter.load_command_line_configuration(args)
        # insert current working directory to the python path to have a correct
        # behaviour
        sys.path.insert(0, os.getcwd())
        if not args or len(args) > 1:
            print linter.help()
            return 
        modname = args[0]
        if modname[-3:] == '.py' or modname.find(os.sep) > -1:
            modname = '.'.join(modpath_from_file(modname))        
        linter.check(modname)
        linter.display_results()

    def generate_config(self, *args, **kwargs):
        """optik callback for sample config file generation
        """
        self.linter.generate_config()
        sys.exit(0)
        
    def disable_all_checkers(self, *args, **kwargs):
        """optik callback for disabling all checkers
        """
        self.linter.disable_all_checkers()
        
    def set_format(self, option, opt_name, value, parser):
        """optik callback for HTML format setting 
        """
        from logilab.pylint.reporters.html import HTMLReporter
        self.linter.reporter = HTMLReporter()
        
    def init_zope(self, option, opt_name, value, parser):
        """optik callback for Zope products initialization
        """
        import Zope
        try:
            # Zope >= 2.6
            Zope.startup()
        except AttributeError:
            pass


if __name__ == "__main__":
    Run(*sys.argv[1:])
